__doc__='''ZCache'''
__version__='0.4'

from Globals import HTMLFile, MessageDialog, package_home
import OFS.SimpleItem, Acquisition, AccessControl.Role
from Persistence import Persistent
from DateTime import DateTime
import OFS.content_types
import sys, marshal, os
import fcntl

def manage_addZCache(self, id, cacheObjectPath, delay, REQUEST=None):
    """Add ZCache Object"""
    i=ZCache(id, cacheObjectPath, delay, REQUEST)
    self._setObject(id,i)
    if REQUEST is not None:
	return self.manage_main(self,REQUEST)

manage_addZCacheForm=HTMLFile('zcacheAdd', globals())

class ZCache(
    OFS.SimpleItem.Item,
    Persistent,
    Acquisition.Implicit,
    AccessControl.Role.RoleManager,
    ):
	"""ZCache object - store self.cacheObjectPath's results in memory, and return it when called, 
     updating the stored value at most every self.delay seconds."""
	meta_type = 'ZCache'

	icon = 'zcache'

	manage=manage_main=HTMLFile('manageZCache', globals())

	manage_options=({'icon':'', 'label':'Properties',
                    'action':'manage_main', 'target':'manage_main'},
				    {'icon':'', 'label':'View',
                    'action':'index_html', 'target':'manage_main'},
					{'icon':'', 'label':'Security',
                    'action':'manage_access', 'target':'manage_main'})
			 
	__ac_permissions__=(
	        ('View management screens', ('manage', 'manage_main', 'manage_access',)),
		('Manage properties', ('manage_makeChanges',)),
		('View', ('index_html',)),
		('Update Cache', ('update',)),
		)
					
    
	def __init__(self, id, cacheObjectPath, delay, REQUEST):
		"""Create a ZCache - pass object to cache and minimum delay between reads (in seconds) """
		self.id	= id
		self.cacheObjectPath = cacheObjectPath # absolute path from BASE0 of object to cache
		self.lastreadtime = 0          #time of last update
		self.inode = 0				   #inode number of cache's storage file
		self.etag = None			   #Etag http header - generated from lastreadtime and inode
		self.delay = delay			   #interval between updates
		self.uid = None                #name of file where cache is kept - must be unique
		self.contenttype = 'application/octet-stream' #content-type MIME header of cache contents
		self._v_cache = None           #cache contents
		self.size = 0                  #size of contents of cache
		self._v_home = os.path.join(INSTANCE_HOME,'var','ZCache') #folder where we keep cache files
		    
	def _savecache(self):
		""" save cache contents to a file on the harddisk """
		jn = os.path.join
		if self.uid == None:
			from whrandom import randint
			self.uid = str(randint(0, 99)) +  str(DateTime().timeTime()) + str(randint(0, 999999)) # Is this the best way?
			
			#cache files are distributed into 10 x 10 folders to speed
			#access when there are a lot.  It may be faster to put them all
			#in one folder, though...
			exists = os.path.exists
					
			dirpath = jn(self._v_home, self.uid[0])
			if not exists(dirpath):
				os.mkdir(dirpath)
		
			dirpath = jn(dirpath, self.uid[1])
			if not exists(dirpath):
				os.mkdir(dirpath)
				
			dirpath = jn(dirpath, self.uid)
			f = open(dirpath, "w") #create file storage
			f.close()
			self.inode = os.stat(dirpath)[1]
			
		f = open(jn(self._v_home, self.uid[0], self.uid[1], self.uid), "w")
		fcntl.flock(f.fileno(), fcntl.LOCK_EX) #block until only this process has this file open
		f.seek(0, 0)
		marshal.dump(self._v_cache, f)
		f.close()
		
	def _loadcache(self):
		""" load cache contents from file on harddisk """
		filepath = os.path.join(self._v_home, self.uid[0], self.uid[1], self.uid)
		if os.path.exists(filepath):
			f = open(filepath, "r")
			fcntl.flock(f.fileno(), fcntl.LOCK_SH) #block until no process is writing to file
			f.seek(0, 0)
			self._v_cache = marshal.load(f)
			f.close()
		
	def index_html(self, REQUEST, RESPONSE):
		"""Return contents of cache, updating if hasn't been updated for more that self.delay seconds"""
		from string import split
		from webdav.common import rfc1123_date
		RESPONSE.setHeader('Etag', self.etag)
		
		match_etag = REQUEST.get_header('If-None-Match', None)
		if (match_etag is not None) and (match_etag != self.etag):
			RESPONSE.setStatus(304)
			return RESPONSE
		
		ms=REQUEST.get_header('If-Modified-Since', None)
		if ms is not None:
			ms=split(ms, ';')[0]
			ms=DateTime(ms).timeTime()
			if self.lastreadtime <= ms:
				RESPONSE.setStatus(304)
				return RESPONSE
		
		now = DateTime().timeTime()
																					
		if (self._v_cache is not None) and (now > self.lastreadtime + self.delay):
			self._v_cache = None
		if self._v_cache is None:
			self._update()
		RESPONSE.setHeader('Last-Modified', rfc1123_date(self.lastreadtime))
		RESPONSE.setHeader('Expires', rfc1123_date(self.lastreadtime + self.delay))
		RESPONSE.setHeader('Content-Length', self.size)
		RESPONSE.setHeader('Content-Type', self.contenttype)
		RESPONSE.write(self._v_cache)
		return ''

	def __str__(self):
		""" return contents of cache when called by <dtml-var> """
		now = DateTime().timeTime()
		if (self._v_cache is not None) and (now > self.lastreadtime + self.delay):
			self._v_cache = None
		if self._v_cache is None:
			self._update()
		return self._v_cache
			
	def __setstate__(self, state):
		""" initialize zcache when it loads from storage """
		Persistent.__setstate__(self, state)
		self._v_home = os.path.join(INSTANCE_HOME,'var','ZCache')
		
		now = DateTime().timeTime()
		if (now <= self.lastreadtime + self.delay):
			self._loadcache()
		else:
			# if we need to update the cache __str__ or index_html will
			# call _update for use - no need to do so ourselves
			# but we need to create _v_ vars
			self._v_cache = None
			
	def _update(self):
		""" Force update of contents of cache """
		
		REQUEST = self.REQUEST
		cacheObject = REQUEST.resolve_url(REQUEST.BASE0 + '/' + self.cacheObjectPath)
		if REQUEST['AUTHENTICATED_USER'].has_permission("View", cacheObject):
			try: newcache = cacheObject(cacheObject.aq_parent, REQUEST)
			except: raise sys.exc_type, sys.exc_value
			if newcache != self._v_cache:
				self._v_cache = newcache
				self.lastreadtime = int(DateTime().timeTime())
				self.contenttype = OFS.content_types.guess_content_type(self.cacheObjectPath, newcache, 'application/octet-stream')[0]
				self.size = len(self._v_cache)
				self.etag = '"' + str(self.inode) + str(self.lastreadtime) + '"'
				self._savecache()
		else:
			raise 'Unauthorized', ('You are not authorized to access the cached object.')
	
	def update(self):
		""" Web interface to _update """
		self._update()
	
	def manage_makeChanges(self, cacheObjectPath, delay, REQUEST=None):
		""" Change properties """
		self.cacheObjectPath = cacheObjectPath
		self.delay = delay
		self._update()
		if REQUEST: return MessageDialog(title  ='Changed %s' % self.__name__,
					 message='%s has been updated' % self.id,
				     action =REQUEST['URL2']+'/manage_main',
					 target ='manage_main')